use base64::{decode, encode};
use std::error::Error;
use std::hash::Hash;
use std::io;
use std::net::{IpAddr, SocketAddr};
use std::time::Duration;
use std::time::Instant;

const KEY_SIZE: usize = 48; //TODO: use VEC instead of array

#[derive(Debug, PartialEq, Eq, Hash, Clone)]
pub enum ECCKey {
    PublicKey(Vec<u8>),
    PrivateKey(Vec<u8>),
}

pub trait Base64Backed {
    fn from_bytes(bytes: Vec<u8>) -> Self;
    fn bytes(&self) -> &Vec<u8>;
    fn from_base64<I: AsRef<str>>(key: I) -> io::Result<Self>
    where
        Self: Sized,
    {
        let key = match decode(key.as_ref()) {
            Ok(key) => key,
            _ => {
                return Err(io::Error::new(
                    io::ErrorKind::InvalidData,
                    "Failed to decode base64",
                ))
            }
        }; /*.map_err(|err| {

           })?;*/
        if key.len() != KEY_SIZE {
            return Err(io::Error::new(
                io::ErrorKind::Other,
                format!(
                    "Mismatched key size. Expected: {}, Got {}",
                    KEY_SIZE,
                    key.len()
                ),
            ));
        }
        Ok(Self::from_bytes(key))
    }

    fn as_base64(&self) -> io::Result<String> {
        Ok(encode(self.bytes()))
    }
}

impl Base64Backed for ECCKey {
    fn bytes(&self) -> &Vec<u8> {
        match self {
            ECCKey::PublicKey(bytes) => &bytes,
            ECCKey::PrivateKey(bytes) => &bytes,
        }
    }

    fn from_bytes(bytes: Vec<u8>) -> ECCKey {
        ECCKey::PublicKey(bytes)
    }
}

impl ECCKey {
    pub fn extract_public_key(&self) -> Option<ECCKey> {
        //TODO: Determine whether Self is a private key and only the return public part
        Some(self.clone())
    }
}

#[derive(Debug, PartialEq, Eq, Hash, Clone)]
pub struct SharedKey(Vec<u8>);

impl Base64Backed for SharedKey {
    fn bytes(&self) -> &Vec<u8> {
        &self.0
    }
    fn from_bytes(bytes: Vec<u8>) -> SharedKey {
        SharedKey(bytes)
    }
}

#[derive(Debug, Builder, PartialEq, Eq, Hash, Clone)]
pub struct Interface {
    pub key: ECCKey,
    pub port: usize,
    pub fwmark: Option<String>,
}

#[derive(Debug, Builder, PartialEq, Eq, Hash, Clone)]
pub struct Peer {
    pub key: ECCKey,
    #[builder(default = "None")]
    pub shared_key: Option<SharedKey>,
    #[builder(default = "None")]
    pub endpoint: Option<SocketAddr>,
    #[builder(default = "Vec::new()")]
    pub allowed_ips: Vec<(IpAddr, u8)>,
    #[builder(default = "None")]
    pub last_handshake: Option<Duration>,
    #[builder(default = "None")]
    pub persistent_keepalive: Option<Duration>,
    #[builder(default = "(0u64,0u64)")]
    pub traffic: (u64, u64),
    #[builder(default = "Instant::now()")]
    pub parsed: Instant,
}

impl PeerBuilder {
    fn validate(&self) -> Result<(), String> {
        if let Some(ref key) = self.key {
            Ok(())
        } else {
            Err("No key supplied".into())
        }
    }

    pub fn is_whole(&self) -> bool {
        self.validate().is_ok()
    }

    pub fn add_allowed_ip(&mut self, ip: (IpAddr, u8)) {
        if let Some(ref mut ips) = &mut self.allowed_ips {
            ips.push(ip);
        } else {
            self.allowed_ips = Some(vec![ip]);
        }
    }

    pub fn add_last_handshake(&mut self, d: Duration) {
        if !self.last_handshake.is_some() {
            self.last_handshake = Some(Some(d));
        } else {
            self.last_handshake = self
                .last_handshake
                .map(|shake| shake.map(|shake| shake + d));
        }
    }

    pub fn add_traffic(&mut self, txrx: (u64, u64)) {
        if let Some(ref mut traffic) = &mut self.traffic {
            traffic.0 += txrx.0;
            traffic.1 += txrx.1;
        } else {
            self.traffic = Some(txrx);
        }
    }
}

pub trait WireguardController {
    fn peers<'a>(&'a mut self) -> io::Result<Box<Iterator<Item = io::Result<Peer>> + 'a>>;

    fn interface(&mut self) -> io::Result<Interface>;

    fn update_peer(&mut self, peer: &Peer) -> io::Result<()>;
}
